# Copyright Red Hat 2017, Jake Hunsaker <jhunsake@redhat.com>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import fnmatch

from soscollector.clusters import Cluster
from getpass import getpass


class ovirt(Cluster):

    packages = ('ovirt-engine', 'rhevm')
    sos_plugins = ['ovirt']

    option_list = [
        ('no-database', False, 'Do not collect a database dump'),
        ('cluster', '', 'Only collect from hosts in this cluster'),
        ('datacenter', '', 'Only collect from hosts in this datacenter'),
        ('no-hypervisors', False, 'Do not collect from hypervisors')
    ]

    def setup(self):
        self.pg_pass = False
        if not self.get_option('no-database'):
            self.conf = self.parse_db_conf()
        self.format_db_cmd()

    def format_db_cmd(self):
        self.dbcmd = '/usr/share/ovirt-engine/dbscripts/engine-psql.sh -c '
        self.dbcmd += '"select host_name from vds_static "'
        if self.get_option('cluster'):
            self.dbcmd += ('" where cluster_id = (select '
                           'cluster_id from cluster where name = \'%s\')"'
                           % self.get_option('cluster'))
        if self.get_option('datacenter'):
            self.dbcmd += ('"where cluster_id = (select cluster_id from '
                           'cluster where storage_pool_id = (select id from '
                           'storage_pool where name = \'%s\')) "'
                           % self.get_option('datacenter'))
        self.log_debug('Query command for ovirt DB set to: %s' % self.dbcmd)

    def get_nodes(self):
        if self.get_option('no-hypervisors'):
            return []
        res = self.exec_master_cmd(self.dbcmd)
        if res['status'] == 0:
            nodes = res['stdout'].splitlines()[2:-1]
            return [n.split('(')[0].strip() for n in nodes]
        else:
            raise Exception('database query failed, return code: %s'
                            % res['status'])

    def set_node_label(self, facts):
        if facts['address'] == self.master.address:
            return 'manager'
        if 'hypervisor' in facts['release']:
            return 'rhvh'
        else:
            return 'rhelh'

    def run_extra_cmd(self):
        if not self.get_option('no-database'):
            return self.collect_database()
        return False

    def parse_db_conf(self):
        conf = {}
        engconf = '/etc/ovirt-engine/engine.conf.d/10-setup-database.conf'
        res = self.exec_master_cmd('cat %s' % engconf)
        if res['status'] == 0:
            config = res['stdout'].splitlines()
        for line in config:
            k = str(line.split('=')[0])
            v = str(line.split('=')[1].replace('"', ''))
            conf[k] = v
        return conf

    def collect_database(self):
        sos_opt = (
                   '-k {plugin}.dbname={db} '
                   '-k {plugin}.dbhost={dbhost} '
                   '-k {plugin}.dbport={dbport} '
                   '-k {plugin}.username={dbuser} '
                   ).format(plugin='postgresql',
                            db=self.conf['ENGINE_DB_DATABASE'],
                            dbhost=self.conf['ENGINE_DB_HOST'],
                            dbport=self.conf['ENGINE_DB_PORT'],
                            dbuser=self.conf['ENGINE_DB_USER']
                            )
        cmd = ('PGPASSWORD={} /usr/sbin/sosreport --name=postgresql '
               '--batch -o postgresql {}'
               ).format(self.conf['ENGINE_DB_PASSWORD'], sos_opt)
        db_sos = self.exec_master_cmd(cmd)
        for line in db_sos['stdout'].splitlines():
            if fnmatch.fnmatch(line, '*sosreport-*tar*'):
                return line.strip()
        self.log_error('Failed to gather database dump')
        return False



# Copyright Red Hat 2017, Jake Hunsaker <jhunsake@redhat.com>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import subprocess

from soscollector.configuration import ClusterOption


class Cluster():

    option_list = []
    packages = ('',)
    sos_options = {}

    def __init__(self, config):
        '''This is the class that cluster profile should subclass in order to
        add support for different clustering technologies and environments to
        sos-collector.

        A profile should at minimum define a package that indicates the node is
        configured for the type of cluster the profile is intended to serve and
        then additionall be able to return a list of enumerated nodes via the
        get_nodes() method
        '''

        self.master = None
        self.config = config
        self.cluster_type = self.__class__.__name__
        self.node_list = None
        sos_plugins = []
        self.logger = logging.getLogger('sos_collector')
        self.console = logging.getLogger('sos_collector_console')
        self.options = []
        self._get_options()

    def _get_options(self):
        '''Loads the options defined by a cluster and sets the default value'''
        for opt in self.option_list:
            option = ClusterOption(name=opt[0], opt_type=opt[1].__class__,
                                   value=opt[1], cluster=self.cluster_type,
                                   description=opt[2])
            self.options.append(option)

    def _fmt_msg(self, msg):
        return '[ %s ] %s' % (self.cluster_type, msg)

    def log_info(self, msg):
        '''Used to print info messages'''
        self.logger.info(self._fmt_msg(msg))
        self.console.info(msg)

    def log_error(self, msg):
        '''Used to print error messages'''
        self.logger.error(self._fmt_msg(msg))
        self.console.error(msg)

    def log_debug(self, msg):
        '''Used to print debug messages'''
        self.logger.debug(self._fmt_msg(msg))
        self.console.debug(self._fmt_msg(msg))

    def get_option(self, option):
        '''This is used to by clusters to check if a cluster option was
        supplied to sos-collector.
        '''
        for opt in self.options:
            if opt.name == option:
                return opt.value
        return False

    def is_installed(self, pkg):
        '''Checks to see if a package is installed'''
        cmd = self.master.host_facts['package_manager']['query'] + pkg
        res = self.exec_master_cmd(cmd)
        if res['status'] == 0:
            return True
        return False

    def exec_master_cmd(self, cmd):
        '''Used to retrieve output from a (master) node in a cluster'''
        if self.config['need_sudo']:
            cmd = "sudo -S %s" % cmd
        if self.config['become_root']:
            cmd = "su -c '%s'" % cmd
        self.logger.debug('Running %s on %s' % (cmd, self.master.address))
        res = self.master.run_command(cmd, get_pty=True)
        if res['stdout']:
            if 'password' in res['stdout'][0].lower():
                res['stdout'].pop(0)
        return res

    def setup(self):
        '''This MAY be used by a cluster to do prep work in case there are
        extra commands to be run even if a node list is given by the user, and
        thus get_nodes() would not be called
        '''
        pass

    def get_sos_path_strip(self, facts):
        '''This calls set_sos_path_strip that is used by clusters to determine
        if we need to remove a particular string from a returned sos path for
        any reason
        '''
        return self.set_sos_path_strip(facts)

    def set_sos_path_strip(self, facts):
        '''This may be overriden by a cluster and used to set
        a string to be stripped from the return sos path if needed.

        For example, on Atomic Host, the sosreport gets written under
        /host/var/tmp in the container, but is available to scp under the
        standard /var/tmp after the container exits.

        If a cluster overrides this, it will need to be known if it needs to be
        sensitive to cluster nodes being Atomic Hosts.
        '''
        if facts['atomic']:
            return '/host'

    def get_cleanup_cmd(self, facts):
        '''This calls set_cleanup_cmd that is used by clusers to determine if
        sos-collector needs to do additional cleanup on a node
        '''
        return self.set_cleanup_cmd(facts)

    def set_cleanup_cmd(self, facts):
        '''This should be overridden by a cluster and used to set an additional
        command to run during cleanup.

        The cluster should return a string containing the full cleanup
        command to run

        If a cluster overrides this, it will need to be known if the the
        cluster needs to be sensitive to cluster nodes being Atomic Hosts.
        '''
        if facts['atomic']:
            return 'docker rm sos-collector-tmp'

    def check_enabled(self):
        '''This may be overridden by clusters

        This is called by sos-collector on each cluster type that exists, and
        is meant to return True when the cluster type matches a criteria
        that indicates that is the cluster type is in use.

        Only the first cluster type to determine a match is run
        '''
        for pkg in self.packages:
            if self.is_installed(pkg):
                return True
        return False

    def get_nodes(self):
        '''This MUST be overridden by a cluster.
        A cluster should use this method to return a list or string that
        contains all the nodes that a report should be collected from
        '''
        pass

    def _get_nodes(self):
        try:
            return self.format_node_list()
        except Exception as e:
            self.logger.error('Failed to get node list: %s' % e)
            raise

    def get_node_label(self, facts):
        '''Used by SosNode() to retrieve the appropriate label from the cluster
        as set by set_node_label() in the cluster profile.
        '''
        return self.set_node_label(facts)

    def set_node_label(self, facts):
        '''This may be overridden by clusters.

        If there is a distinction between masters and nodes, or types of nodes,
        then this can be used to label the sosreport archive differently.
        '''
        return ''

    def modify_sos_cmd(self):
        '''This is used to modify the sosreport command run on the nodes.
        By default, sosreport is run without any options, using this will
        allow the profile to specify what plugins to run or not and what
        options to use.

        This will NOT override user supplied options.
        '''
        if self.sos_plugins:
            for plug in self.sos_plugins:
                if plug not in self.config['sos_cmd']:
                    self.config['enable_plugins'].append(plug)
        if self.sos_options:
            for opt in self.sos_options:
                if not any(opt in o for o in self.config['plugin_options']):
                    option = '%s=%s' % (opt, self.sos_options[opt])
                    self.config['plugin_options'].append(option)

    def format_node_list(self):
        '''Format the returned list of nodes from a cluster into a known
        format. This being a list that contains no duplicates
        '''
        try:
            nodes = self.get_nodes()
        except Exception as e:
            self.log_error('\n%s failed to enumerate nodes: %s'
                           % (self.cluster_type, e))
            raise
        if isinstance(nodes, list):
            node_list = [n.strip() for n in nodes if n]
            node_list = list(set(nodes))
        if isinstance(nodes, str):
            node_list = [n.split(',').strip() for n in nodes]
            node_list = list(set(nodes))
        for node in node_list:
            if node.startswith(('-', '_', '(', ')', '[', ']', '/', '\\')):
                node_list.remove(node)
        return node_list



# Copyright Red Hat 2017, Jake Hunsaker <jhunsake@redhat.com>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import fnmatch
import logging
import paramiko
import re
import shutil
import socket
import subprocess
import sys
import time

from distutils.version import LooseVersion
from subprocess import Popen, PIPE


class SosNode():

    def __init__(self, address, config, force=False, load_facts=True):
        self.address = address.strip()
        self.local = False
        self.hostname = None
        self.config = config
        self.sos_path = None
        self.retrieved = False
        self.host_facts = {'address': address}
        self.sos_info = {
            'version': None,
            'enabled': [],
            'disabled': [],
            'options': []
        }
        filt = ['localhost', '127.0.0.1', self.config['hostname']]
        self.logger = logging.getLogger('sos_collector')
        self.console = logging.getLogger('sos_collector_console')
        if self.address not in filt or force:
            self.connected = self.open_ssh_session()
        else:
            self.connected = True
            self.local = True
        if self.connected and load_facts:
            self.get_hostname()
            self.load_host_facts()
            self._load_sos_info()

    def _fmt_msg(self, msg):
        return '{:<{}} : {}'.format(self._hostname,
                                    self.config['hostlen'],
                                    msg)

    def file_exists(self, fname):
        '''Checks for the presence of fname on the remote node'''
        if not self.local:
            sftp = self.client.open_sftp()
            try:
                sftp.stat(fname)
                return True
            except Exception:
                return False
        else:
            try:
                os.stat(fname)
                return True
            except Exception:
                return False

    @property
    def _hostname(self):
        return self.hostname if self.hostname else self.address

    def log_info(self, msg):
        '''Used to print and log info messages'''
        self.logger.info(' %s: %s' % (self._hostname, msg))
        self.console.info(self._fmt_msg(msg))

    def log_error(self, msg):
        '''Used to print and log error messages'''
        self.logger.error(' %s: %s' % (self._hostname, msg))
        self.console.info(self._fmt_msg(msg))

    def log_debug(self, msg):
        '''Used to print and log debug messages'''
        self.logger.debug(' %s: %s' % (self._hostname, msg))
        self.console.debug(self._fmt_msg(msg))

    def get_hostname(self):
        '''Get the node's hostname'''
        sout = self.run_command('hostname')
        self.hostname = sout['stdout'].strip()
        self.logger.debug(
            'Hostname for %s set to %s' % (self.address, self.hostname))

    def _format_cmd(self, cmd):
        '''If we need to provide a sudo or root password to a command, then
        here we prefix the command with the correct bits
        '''
        if self.config['become_root']:
            return "su -c '%s'" % cmd
        if self.config['need_sudo']:
            return "sudo -S %s" % cmd
        return cmd

    def _fmt_output(self, stdout=None, stderr=None, rc=0):
        '''Formats the returned output from a command into a dict'''
        c = {}
        c['status'] = rc
        if isinstance(stdout, (bytes, unicode)):
            stdout = [str(stdout)]
        if isinstance(stderr, (bytes, unicode)):
            stderr = [str(stderr)]
        if stdout:
            stdout = ''.join(s for s in stdout) or True
        if stderr:
            stderr = ' '.join(s for s in stderr) or False
        res = {'status': rc,
               'stdout': stdout,
               'stderr': stderr}
        return res

    def _load_sos_info(self):
        '''Queries the node for information about the installed version of sos
        '''
        prefix = self.set_sos_prefix()
        cmd = prefix + self.host_facts['package_manager']['query'] + 'sos'
        res = self.run_command(cmd)
        if res['status'] == 0:
            ver = res['stdout'].splitlines()[-1].split('-')[1]
            self.sos_info['version'] = ver
            self.log_debug('sos version is %s' % self.sos_info['version'])
        else:
            self.log_error('sos is not installed on this node')
            self.connected = False
            return False
        cmd = prefix + 'sosreport -l'
        self.sosinfo = self.run_command(cmd)
        if self.sosinfo['status'] == 0:
            ENABLED = 'The following plugins are currently enabled:'
            DISABLED = 'The following plugins are currently disabled:'
            OPTIONS = 'The following plugin options are available:'
            PROFILES = 'Profiles:'

            enablereg = ENABLED + '(.*?)' + DISABLED
            disreg = DISABLED + '(.*?)' + OPTIONS
            optreg = OPTIONS + '(.*?)' + PROFILES
            proreg = PROFILES + '(.*?)' + '\n\n'

            self.sos_info['enabled'] = self._regex_sos_help(enablereg)
            self.sos_info['disabled'] = self._regex_sos_help(disreg)
            self.sos_info['options'] = self._regex_sos_help(optreg)
            self.sos_info['profiles'] = self._regex_sos_help(proreg, True)

    def _regex_sos_help(self, regex, is_list=False):
        res = []
        for result in re.findall(regex, self.sosinfo['stdout'], re.S):
            for line in result.splitlines():
                if not is_list:
                    try:
                        res.append(line.split()[0])
                    except Exception:
                        pass
                else:
                    r = line.split(',')
                    res.extend(p.strip() for p in r if p.strip())
        return res

    def check_sos_version(self, ver):
        '''Checks to see if the sos installation on the node is AT LEAST the
        given ver. This means that if the installed version is greater than
        ver, this will still return True
        '''
        return LooseVersion(self.sos_info['version']) >= ver

    def run_command(self, cmd, timeout=180, get_pty=False):
        '''Runs a given cmd, either via the SSH session or locally'''
        if 'atomic' in cmd:
            get_pty = True
        if not self.local:
            now = time.time()
            sin, sout, serr = self.client.exec_command(cmd, timeout=timeout,
                                                       get_pty=get_pty)
            if self.config['become_root']:
                sin.write(self.config['root_password'] + '\n')
                sin.flush()
            if self.config['need_sudo']:
                sin.write(self.config['sudo_pw'] + '\n')
                sin.flush()
            while time.time() < now + timeout:
                if sout.channel.exit_status_ready():
                    rc = sout.channel.recv_exit_status()
                    return self._fmt_output(sout, serr, rc)
            else:
                raise socket.timeout
        else:
            proc = Popen(cmd, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            stdout, stderr = proc.communicate()
            if self.config['become_root']:
                proc.communicate(input=self.config['root_password'] + '\n')
            if self.config['need_sudo']:
                proc.communicate(input=self.config['sude_pw'] + '\n')
            rc = proc.returncode
            if stdout:
                sout = (stdout or True)
            else:
                sout = None
            return self._fmt_output(stdout=sout, stderr=stderr, rc=rc)

    def sosreport(self):
        '''Run a sosreport on the node, then collect it'''
        self.finalize_sos_cmd()
        self.logger.debug('Running sosreport on %s' % self.address)
        path = self.execute_sos_command()
        if path:
            self.finalize_sos_path(path)
        else:
            self.log_error('Unable to determine path of sos archive')
        if self.sos_path:
            self.retrieved = self.retrieve_sosreport()
        self.cleanup()

    def open_ssh_session(self):
        '''Create the persistent ssh session we use on the node'''
        try:
            msg = ''
            self.client = paramiko.SSHClient()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.client.load_system_host_keys()
            if not self.config['password']:
                self.logger.debug(
                    'Opening passwordless session to %s' % self.address)
                self.client.connect(self.address,
                                    username=self.config['ssh_user'],
                                    timeout=15)
            else:
                self.logger.debug(
                    'Opening session to %s with password' % self.address)
                self.client.connect(self.address,
                                    username=self.config['ssh_user'],
                                    password=self.config['password'],
                                    timeout=15)
            self.logger.debug('%s successfully connected' % self._hostname)
            return True
        except paramiko.AuthenticationException:
            if not self.config['password']:
                self.log_error('Authentication failed. SSH keys installed?')
            else:
                self.log_error('Authentication failed. Incorrect password.')
        except paramiko.BadAuthenticationType:
            self.log_error('Bad authentication type. The node rejected the '
                           'authentication attempt.')
        except paramiko.BadHostKeyException:
            self.log_error('Provided key was rejected by remote SSH client.'
                           ' Check ~/.ssh/known_hosts.')
        except socket.gaierror as err:
            if err.errno == -2:
                self.log_error('Provided hostname did not resolve.')
            else:
                self.log_error('Socket error trying to connect: %s' % err)
        except Exception as e:
            self.log_error('Exception caught while trying to connect: %s' % e)
        raise

    def close_ssh_session(self):
        '''Handle closing the SSH session'''
        if self.local:
            return True
        try:
            self.client.close()
            self.connected = False
            return True
        except Exception as e:
            self.log_error('Error closing SSH session: %s' % e)
            return False

    def load_host_facts(self):
        '''Obtain information about the node which can be referneced by
        clusters to change the sosreport command'''
        self.get_release()
        self.set_package_manager()
        self.logger.debug('%s facts found to be %s'
                          % (self._hostname, self.host_facts))

    def set_sos_prefix(self):
        '''Sets a prefix to any sos related commands run on the node.
        Currently, only checks for if the node is an Atomic Host, in which case
        use the specified container image to run all sos commands in
        '''
        prefix = ''
        if self.host_facts['atomic']:
            cmd = 'atomic run --name=sos-collector-tmp --replace '
            img = self.config['image']
            prefix = '%s %s ' % (cmd, img)
        return prefix

    def get_release(self):
        '''Determine the distribution that we're running on.
        For our intents, any Red Hat family distribution or derivitive is going
        to be listed as 'Red Hat'
        '''
        release = 'Unknown'
        self.host_facts['release'] = release
        self.host_facts['atomic'] = False
        try:
            if self.file_exists('/etc/redhat-release'):
                relfile = '/etc/redhat-release'
            else:
                relfile = '/etc/os-release'
            res = self.run_command('cat ' + relfile)
            if len(res['stdout'].splitlines()) > 2:
                for line in res['stdout'].splitlines():
                    if line.startswith('NAME'):
                        release = line.split('=')[1].lower().strip('"')
            else:
                release = res['stdout'].lower()
            self.host_facts['release'] = release
            rh = ['fedora', 'centos', 'red hat']
            if any(rel in release for rel in rh):
                self.host_facts['distro'] = 'Red Hat'
                self.config['image'] = ('registry.access.redhat.com/rhel7/'
                                        'support-tools ')
            self.host_facts['atomic'] = 'atomic' in release
        except Exception as e:
            self.log_error(e)

    def set_package_manager(self):
        '''Based on the distribution of the node, set the package manager to
        use for checking system installations'''
        self.host_facts['package_manager'] = None
        if self.host_facts['distro'] == 'Red Hat':
            self.host_facts['package_manager'] = {'name': 'rpm',
                                                  'query': 'rpm -q '
                                                  }

    def _plugin_exists(self, plugin):
        '''Verifies if the given plugin exists on the node'''
        return any(plugin in s for s in [self.sos_info['enabled'],
                                         self.sos_info['disabled']])

    def _check_enabled(self, plugin):
        '''Checks to see if the plugin is default enabled on node'''
        return plugin in self.sos_info['enabled']

    def _check_disabled(self, plugin):
        '''Checks to see if the plugin is default disabled on node'''
        return plugin in self.sos_info['disabled']

    def _plugin_option_exists(self, opt):
        '''Attempts to verify that the given option is available on the node.
        Note that we only get available options for enabled plugins, so if a
        plugin has been force-enabled we cannot validate if the plugin option
        is correct or not'''
        plug = opt.split('.')[0]
        if not self._plugin_exists(plug):
            return False
        if (self._check_disabled(plug) and
                plug not in self.config['enable_plugins']):
            return False
        if self._check_enabled(plug):
            return opt in self.sos_info['options']
        # plugin exists, but is normally disabled. Assume user knows option is
        # valid when enabling the plugin
        return True

    def _fmt_sos_opt_list(self, opts):
        '''Returns a comma delimited list for sos plugins that are confirmed
        to exist on the node'''
        return ','.join(o for o in opts if self._plugin_exists(o))

    def finalize_sos_cmd(self):
        '''Use host facts and compare to the cluster type to modify the sos
        command if needed'''
        self.sos_cmd = self._format_cmd(self.config['sos_cmd'])
        prefix = self.set_sos_prefix()
        if prefix:
            self.sos_cmd = prefix + self.sos_cmd

        if self.config['sos_opt_line']:
            self.sos_cmd += self.config['sos_opt_line']
            return True

        label = self.determine_sos_label()
        if label:
            self.sos_cmd = '%s %s ' % (self.sos_cmd, label)

        if self.config['only_plugins']:
            only = self._fmt_sos_opt_list(self.config['only_plugins'])
            if only:
                self.sos_cmd += '--only-plugins=%s ' % only
            return True

        if self.config['skip_plugins']:
            # only run skip-plugins for plugins that are enabled
            skip = [o for o in self.config['skip_plugins']
                    if self._check_enabled(o)]
            skipln = self._fmt_sos_opt_list(skip)
            if skipln:
                self.sos_cmd += '--skip-plugins=%s ' % skipln

        if self.config['enable_plugins']:
            # only run enable for plugins that are disabled
            opts = [o for o in self.config['enable_plugins']
                    if o not in self.config['skip_plugins']
                    and self._check_disabled(o)]
            enable = self._fmt_sos_opt_list(opts)
            if enable:
                self.sos_cmd += '--enable-plugins=%s ' % enable

        if self.config['plugin_options']:
            opts = [o for o in self.config['plugin_options']
                    if self._plugin_exists(o.split('.')[0])
                    and self._plugin_option_exists(o.split('=')[0])]
            if opts:
                self.sos_cmd += '-k %s' % ','.join(o for o in opts)

        self.log_debug('final sos command set to %s' % self.sos_cmd)

    def determine_sos_label(self):
        '''Determine what, if any, label should be added to the sosreport'''
        label = ''
        label += self.config['cluster'].get_node_label(self.host_facts)

        if self.config['label']:
            label += ('%s' % self.config['label'] if not label
                      else '-%s' % self.config['label'])

        if not label:
            return None

        if self.check_sos_version('3.6'):
            lcmd = '--label'
        else:
            lcmd = '--name'
            label = '%s-%s' % (self.address.split('.')[0], label)
        return '%s=%s' % (lcmd, label)

    def finalize_sos_path(self, path):
        '''Use host facts to determine if we need to change the sos path
        we are retrieving from'''
        pstrip = self.config['cluster'].get_sos_path_strip(self.host_facts)
        if pstrip:
            path = path.replace(pstrip, '')
        path = path.split()[0]
        self.logger.debug('Final sos path for %s: %s' % (self.address, path))
        self.sos_path = path
        self.archive = path.split('/')[-1]

    def determine_sos_error(self, rc, stdout):
        if rc == -1:
            return 'sosreport process received SIGKILL on node'
        if rc == 127:
            return 'sosreport terminated unexpectedly. Check disk space'
        if len(stdout) > 0:
            return stdout.split('\n')[0:1]
        else:
            return 'sos exited with code %s' % rc

    def execute_sos_command(self):
        '''Run sosreport and capture the resulting file path'''
        self.logger.info('Running sosreport on %s' % self.address)
        self.log_info("Generating sosreport...")
        try:
            path = False
            res = self.run_command(self.sos_cmd,
                                   timeout=self.config['timeout'],
                                   get_pty=True)
            if res['status'] == 0:
                for line in res['stdout'].splitlines():
                    if fnmatch.fnmatch(line, '*sosreport-*tar*'):
                        path = line.strip()
            else:
                err = self.determine_sos_error(res['status'], res['stdout'])
                self.log_debug("Error running sosreport. rc = %s msg = %s"
                               % (res['status'], res['stdout'] or
                                  res['stderr']))
                self.log_error('Error running sosreport: %s' % err)
            return path
        except socket.timeout:
            self.log_error('Timeout exceeded')
        except Exception as e:
            self.log_error('Error running sosreport: %s' % e)

    def retrieve_sosreport(self):
        '''Collect the sosreport archive from the node'''
        if self.sos_path:
            if self.config['need_sudo'] or self.config['become_root']:
                try:
                    self.make_archive_readable(self.sos_path)
                except Exception:
                    self.log_error('Failed to make archive readable')
                    return False
            self.logger.info('Retrieving sosreport from %s' % self.address)
            self.log_info('Retrieving sosreport...')
            try:
                dest = self.config['tmp_dir'] + '/' + self.archive
                if not self.local:
                    sftp = self.client.open_sftp()
                    sftp.get(self.sos_path, dest)
                    sftp.close()
                else:
                    shutil.move(self.sos_path, dest)
                self.retrieved = True
                self.log_info('Successfully collected sosreport')
                return True
            except Exception as err:
                msg = 'Failed to retrieve sosreport from %s, error: %s'
                self.logger.error(msg % (self.address, err))
                self.log_error('Failed to retrieve sosreport. %s' % err)
                return False
        else:
            # sos sometimes fails but still returns a 0 exit code
            if self.stderr.read():
                e = self.stderr.read()
            else:
                e = [x.strip() for x in self.stdout.readlines() if x.strip][-1]
            self.logger.error(
                'Failed to run sosreport on %s: %s' % (self.address, e))
            self.log_error('Failed to run sosreport. %s' % e)
            return False

    def remove_sos_archive(self):
        '''Remove the sosreport archive from the node, since we have
        collected it and it would be wasted space otherwise'''
        try:
            cmd = self._format_cmd("rm -f %s" % self.sos_path)
            res = self.run_command(cmd)
        except Exception as e:
            self.log_error('Failed to remove sosreport on host: %s' % e)

    def cleanup(self):
        '''Remove the sos archive from the node once we have it locally'''
        self.remove_sos_archive()
        cleanup = self.config['cluster'].get_cleanup_cmd(self.host_facts)
        if cleanup:
            sin, sout, serr = self.client.exec_command(cleanup, timeout=15)

    def collect_extra_cmd(self, filename):
        '''Collect the file created by a cluster outside of sos'''
        try:
            if self.config['need_sudo'] or self.config['become_root']:
                try:
                    self.make_archive_readable(filename)
                except Exception:
                    self.console.error('Unable to make extra data readable')
                    return False
            dest = self.config['tmp_dir'] + '/' + filename.split('/')[-1]
            if not self.local:
                sftp = self.client.open_sftp()
                sftp.get(filename, dest)
                sftp.close()
            else:
                shutil.move(filename, dest)
            return True
        except Exception as e:
            msg = 'Error collecting additional data from master: %s' % e
            self.console.error(msg)
            return False

    def make_archive_readable(self, filepath):
        '''Used to make the given archive world-readable, which is slightly
        better than changing the ownership outright.

        This is only used when we're not connecting as root.
        '''
        cmd = self._format_cmd('chmod +r %s' % filepath)
        res = self.run_command(cmd, timeout=10, get_pty=True)
        if res['status'] == 0:
            return True
        else:
            msg = "Exception while making %s readable. Return code was %s"
            self.logger.error(msg % (filepath, res['status']))
            raise Exception



# Copyright Red Hat 2017, Jake Hunsaker <jhunsake@redhat.com>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import re
import six
import socket


class Configuration(dict):
    """ Dict subclass that is used to handle configuration information
    needed by both SosCollector and the SosNode classes """

    def __init__(self, args=None):
        self.args = args
        self.set_defaults()
        self.parse_config()
        self.parse_options()
        self.check_user_privs()
        self.parse_node_strings()

    def set_defaults(self):
        self['sos_mod'] = {}
        self['master'] = ''
        self['strip_sos_path'] = ''
        self['ssh_port'] = '22'
        self['ssh_user'] = 'root'
        self['sos_cmd'] = '/usr/sbin/sosreport --batch '
        self['no_local'] = False
        self['tmp_dir'] = None
        self['out_dir'] = '/var/tmp/'
        self['nodes'] = None
        self['debug'] = False
        self['tmp_dir_created'] = False
        self['cluster_type'] = None
        self['cluster'] = None
        self['password'] = False
        self['label'] = None
        self['case_id'] = None
        self['timeout'] = 300
        self['all_logs'] = False
        self['alloptions'] = False
        self['no_pkg_check'] = False
        self['hostname'] = socket.gethostname()
        ips = [i[4][0] for i in socket.getaddrinfo(socket.gethostname(), None)]
        self['ip_addrs'] = list(set(ips))
        self['cluster_options'] = ''
        self['image'] = None
        self['skip_plugins'] = []
        self['enable_plugins'] = []
        self['plugin_options'] = []
        self['only_plugins'] = []
        self['list_options'] = False
        self['hostlen'] = len(self['master']) or len(self['hostname'])
        self['need_sudo'] = False
        self['sudo_pw'] = ''
        self['become_root'] = False
        self['root_password'] = ''
        self['threads'] = 4
        self['compression'] = ''
        self['verify'] = False
        self['chroot'] = ''
        self['sysroot'] = ''
        self['sos_opt_line'] = ''

    def parse_node_strings(self):
        '''
        Parses the given --nodes option(s) to properly format the regex
        list that we use. We cannot blindly split on ',' chars since it is a
        valid regex character, so we need to scan along the given strings and
        check at each comma if we should use the preceeding string by itself
        or not, based on if there is a valid regex at that index.
        '''
        if not self['nodes']:
            return
        nodes = []
        if not isinstance(self['nodes'], list):
            self['nodes'] = [self['nodes']]
        for node in self['nodes']:
            idxs = [i for i, m in enumerate(node) if m == ',']
            idxs.append(len(node))
            start = 0
            pos = 0
            for idx in idxs:
                try:
                    pos = idx
                    reg = node[start:idx]
                    re.compile(reg)
                    nodes.append(reg.lstrip(','))
                    start = idx
                except re.error:
                    continue
            if pos != len(node):
                nodes.append(node[pos+1:])
        self['nodes'] = nodes

    def parse_config(self):
        for k in self.args:
            if self.args[k]:
                self[k] = self.args[k]

    def parse_cluster_options(self):
        opts = []
        if self['cluster_options']:
            for opt in self['cluster_options'].split(','):
                cluster = opt.split('.')[0]
                name = opt.split('.')[1].split('=')[0]
                try:
                    value = opt.split('=')[1]
                except IndexError:
                    # conversion to boolean is handled during validation
                    value = 'True'
                opts.append(
                    ClusterOption(name, value, value.__class__, cluster)
                )
        self['cluster_options'] = opts

    def parse_options(self):
        self.parse_cluster_options()
        for opt in ['skip_plugins', 'enable_plugins', 'plugin_options',
                    'only_plugins']:
            if self[opt]:
                opts = []
                if isinstance(self[opt], six.string_types):
                    self[opt] = [self[opt]]
                for option in self[opt]:
                    opts += option.split(',')
                self[opt] = opts

    def check_user_privs(self):
        if not self['ssh_user'] == 'root':
            self['need_sudo'] = True


class ClusterOption():
    '''Used to store/manipulate options for cluster profiles.'''

    def __init__(self, name, value, opt_type, cluster, description=None):
        self.name = name
        self.value = value
        self.opt_type = opt_type
        self.cluster = cluster
        self.description = description
